#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""Functions used as examples, for testing, or in simple situations"""
import json
import pickle
from typing import Any


def readLines(fileName: str) -> list:
    """
    Simple function to read all lines of a file.
    """
    with open(fileName, "r") as f:
        lines = f.read().split()
    return lines


def createRange(low: int, high: int, step: int) -> range:
    """
    Create and return a range.

    Inputs:
        low: the lower boundary of the range
        high: the higher boundary of the range (excluisve)
        step: Step size between items

    Returns:
        range: iterator
    """
    return range(low, high, step)


def createMultiOut(v1: Any = 1, v2: Any = 7) -> tuple:
    """
    Create two outputs with different values.

    Inputs:
        v1: value 1
        v2: value 2

    Returns:
        tuple of two values
    """
    out1 = v1
    out2 = v2
    return out1, out2


def getMax(v1: float = 1.0, v2: float = 7.0) -> float:
    """
    Return the largest float of two different values.

    Inputs:
        v1: value 1
        v2: value 2

    Returns:
        max(v1, v2)
    """
    return max(v1, v2)


def string2json(string: str, pickle_flag: bool = False) -> list:
    """
    Simple function to convert a string to a JSON object

    Inputs:
        string: string containing a JSON representation
        pickle_flag: False, whether output should be pickled
    """
    if not string:
        string = '""'

    try:
        if not pickle_flag:
            return json.loads(string)
        else:
            return pickle.dumps(json.loads(string))
    except json.decoder.JSONDecodeError:
        return [] if not pickle_flag else pickle.dumps([])



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received block_a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
# pylint: disable=possibly-used-before-assignment
"""
Module tests the shared memory primitive.
"""
import pickle
import sys
import unittest

if sys.version_info >= (3, 8):
    from dlg.shared_memory import DlgSharedMemory, _MAXNAMELENGTH


@unittest.skipIf(sys.version_info < (3, 8), "Shared memory does not work < python 3.8")
class TestSharedMemory(unittest.TestCase):
    """
    Tests the shared memory primitive for atypical and typical uses.
    """

    def test_resize_grow(self):
        """
        The size should grow when requested
        """
        block_a = DlgSharedMemory("A")
        old_size = block_a.size
        block_a.resize(block_a.size * 2)
        self.assertEqual(block_a.size, old_size * 2)
        block_a.close()
        block_a.unlink()

    def test_resize_shrink(self):
        """
        Size should shrink when requested
        """
        block_a = DlgSharedMemory("A")
        old_size = block_a.size
        block_a.resize(old_size // 2)
        self.assertEqual(block_a.size, old_size // 2)
        block_a.close()
        block_a.unlink()

    def test_open_new(self):
        """
        When opening a new block by name, that name should be respected
        """
        block_a = DlgSharedMemory("A")
        self.assertEqual(block_a.name, "/A")
        block_a.close()
        block_a.unlink()

    def test_open_existing(self):
        """
        When opening an existing but still linked block, data should persist
        """
        block_a = DlgSharedMemory("A")
        data = pickle.dumps(3)
        block_a.buf[0: len(data)] = data
        block_a.resize(len(data))
        old_size = block_a.size
        block_a.close()
        block_b = DlgSharedMemory("A")
        self.assertEqual(block_b.size, old_size)
        self.assertEqual(block_b.name, "/A")
        self.assertEqual(block_b.buf[:], data)
        block_b.close()
        block_b.unlink()

    def test_open_random(self):
        """
        It should be possible to open a randomly named block (although this must be deliberate)
        """
        block_a = DlgSharedMemory(None)
        self.assertIsNotNone(block_a.name)
        block_a.close()
        block_a.unlink()

    def test_open_twice(self):
        """
        It should be possible to open a shared memory block simultaneously and therefore offer a
        view into the same memory buffer.
        This does bring to light that this primitive is not thread-safe by default.
        """
        block_a = DlgSharedMemory("A")
        block_b = DlgSharedMemory("A")
        self.assertEqual(block_a.buf, block_b.buf)
        block_a.close()
        block_b.close()
        block_a.unlink()

    def test_unlink_twice(self):
        """
        It should not be possible to unlink a block twice.
        """
        block_a = DlgSharedMemory("A")
        block_b = DlgSharedMemory("A")
        block_a.close()
        block_b.close()
        block_a.unlink()
        with self.assertWarns(RuntimeWarning):
            block_b.unlink()

    def test_long_name(self):
        """
        There is a maximum name size which should be respected
        """
        filename = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        block_a = DlgSharedMemory(filename)
        self.assertEqual(block_a.name, "/" + filename[: _MAXNAMELENGTH - 1])
        block_a.close()
        block_a.unlink()

    def test_start_slash(self):
        """
        If a user attempts to make a block with a name beginning with /, it should be ignored
        (rather than having // as a name)
        """
        filename = "/memory"
        block_a = DlgSharedMemory(filename)
        self.assertEqual(block_a.name, filename)
        block_a.close()
        block_a.unlink()



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

import base64
import contextlib
import io
import time
import os, unittest
import pickle
import random
import shutil
import sqlite3
import string
import sys
import tempfile
import subprocess

from dlg import droputils
from dlg.common.reproducibility.constants import ReproducibilityFlags
from dlg.ddap_protocol import DROPStates, ExecutionMode, AppDROPStates
from dlg.apps.app_base import AppDROP, BarrierAppDROP, InputFiredAppDROP
from dlg.data.drops.data_base import NullDROP
from dlg.data.drops.container import ContainerDROP
from dlg.data.drops.rdbms import RDBMSDrop
from dlg.data.drops.memory import InMemoryDROP, SharedMemoryDROP
from dlg.data.drops.directory import DirectoryDROP
from dlg.data.drops.file import FileDROP
from dlg.droputils import DROPWaiterCtx
from dlg.exceptions import InvalidDropException, ErrorManagerCaughtException, \
    InvalidDROPState
from dlg.apps.simple import Branch
from dlg.apps.simple import NullBarrierApp, SleepAndCopyApp

from test.dlg_engine_testutils import run_errormanagement_exception_test

try:
    from crc32c import crc32c
except:
    from binascii import crc32

ONE_MB = 1024**2


def _start_ns_thread(ns_daemon):
    ns_daemon.requestLoop()


def isContainer(drop):
    # return isinstance(drop, ContainerDROP)
    # A Pyro-friendly way to check for a ContainerDROP is to see if
    # invoking its 'children' attribute fails or not
    try:
        drop.children
        return True
    except AttributeError:
        return False


class SumupContainerChecksum(BarrierAppDROP):
    """
    A dummy BarrierAppDROP that recursively sums up the checksums of
    all the individual DROPs it consumes, and then stores the final
    result in its output DROP
    """

    def run(self):
        crcSum = 0
        for inputDrop in self.inputs:
            if inputDrop.status == DROPStates.COMPLETED:
                if inputDrop.checksum:
                    crcSum += inputDrop.checksum
        outputDrop = self.outputs[0]
        outputDrop.write(str(crcSum).encode("utf8"))


class TestDROP(unittest.TestCase):
    """
    DataDROP related unit tests
    """

    def setUp(self):
        """
        library-specific setup
        """
        self._test_drop_sz = 16  # MB
        self._test_block_sz = 2  # MB
        self._test_num_blocks = self._test_drop_sz // self._test_block_sz
        self._test_block = os.urandom(self._test_block_sz * ONE_MB)

    def tearDown(self):
        shutil.rmtree("/tmp/daliuge_tfiles", True)

    def test_NullDROP(self):
        """
        Check that the NullDROP is usable for testing
        """
        a = NullDROP("A", "A", expectedSize=5)
        a.write(b"1234")
        a.write(b"5")
        allContents = droputils.allDropContents(a)
        self.assertFalse(allContents)

    def test_write_FileDROP(self):
        """
        Test a FileDROP and a simple AppDROP (for checksum calculation)
        """
        self._test_write_withDropType(FileDROP)

    def test_write_InMemoryDROP(self):
        """
        Test an InMemoryDROP and a simple AppDROP (for checksum calculation)
        """
        self._test_write_withDropType(InMemoryDROP)

    def test_dynamic_write_InMemoryDROP(self):
        """
        Test an InMemoryDROP and a simple AppDROP (for checksum calculation)
        """
        self._test_dynamic_write_withDropType(InMemoryDROP)

    @unittest.skipIf(
        sys.version_info < (3, 8), "Shared memory does nt work < python 3.8"
    )
    def test_write_SharedMemoryDROP(self):
        """
        Test a SharedMemoryDROP with simple AppDROP (for checksum calculation)
        """
        self._test_write_withDropType(SharedMemoryDROP)

    @unittest.skipIf(
        sys.version_info < (3, 8), "Shared memory does nt work < python 3.8"
    )
    def test_dynamic_write_SharedMemoryDROP(self):
        """
        Test a SharedMemoryDROP with simple AppDROP (for checksum calculation)
        """
        self._test_dynamic_write_withDropType(SharedMemoryDROP)

    def _test_write_withDropType(self, dropType):
        """
        Test an AbstractDROP and a simple AppDROP (for checksum calculation)
        """
        a = dropType("oid:A", "uid:A", expectedSize=self._test_drop_sz * ONE_MB)
        b = SumupContainerChecksum("oid:B", "uid:B")
        c = InMemoryDROP("oid:C", "uid:C")
        b.addInput(a)
        b.addOutput(c)

        test_crc = 0
        with DROPWaiterCtx(self, c):
            for _ in range(self._test_num_blocks):
                a.write(self._test_block)
                test_crc = crc32c(self._test_block, test_crc)

        # Read the checksum from c
        cChecksum = int(droputils.allDropContents(c))

        self.assertNotEqual(a.checksum, 0)
        self.assertEqual(a.checksum, test_crc)
        self.assertEqual(cChecksum, test_crc)

    def _test_dynamic_write_withDropType(self, dropType):
        """
        Test an AbstractDROP and a simple AppDROP (for checksum calculation)
        without an expected drop size (for app compatibility and not
        recommended in production)
        """
        # NOTE: use_staging required for multiple writes to plasma drops
        a = dropType("oid:A", "uid:A", expectedSize=-1, use_staging=True)
        b = SumupContainerChecksum("oid:B", "uid:B")
        c = InMemoryDROP("oid:C", "uid:C")
        b.addInput(a)
        b.addOutput(c)

        test_crc = 0
        with DROPWaiterCtx(self, c):
            for _ in range(self._test_num_blocks):
                a.write(self._test_block)
                test_crc = crc32c(self._test_block, test_crc)
            a.setCompleted()

        # Read the checksum from c
        cChecksum = int(droputils.allDropContents(c))

        self.assertNotEqual(a.checksum, 0)
        self.assertEqual(a.checksum, test_crc)
        self.assertEqual(cChecksum, test_crc)

    def test_no_write_to_file_drop(self):
        """Check that FileDrops can be *not* written"""
        a = FileDROP("a", "a")
        b = SleepAndCopyApp("b", "b")
        c = InMemoryDROP("c", "c")
        a.addConsumer(b)
        b.addOutput(c)
        with DROPWaiterCtx(self, c):
            a.setCompleted()
        self.assertEqual(droputils.allDropContents(c), b"")

    def test_simple_chain(self):
        """
        Simple test that creates a pipeline-like chain of commands.
        In this case we simulate a pipeline that does this, holding
        each intermediate result in memory:

        cat someFile | grep 'a' | sort | rev
        """

        class GrepResult(BarrierAppDROP):
            def initialize(self, **kwargs):
                super(GrepResult, self).initialize(**kwargs)
                self._substring = kwargs["substring"]

            def run(self):
                drop = self.inputs[0]
                output = self.outputs[0]
                allLines = io.BytesIO(droputils.allDropContents(drop)).readlines()
                for line in allLines:
                    if self._substring in line:
                        output.write(line)

        class SortResult(BarrierAppDROP):
            def run(self):
                drop = self.inputs[0]
                output = self.outputs[0]
                sortedLines = io.BytesIO(droputils.allDropContents(drop)).readlines()
                sortedLines.sort()
                for line in sortedLines:
                    output.write(line)

        class RevResult(BarrierAppDROP):
            def run(self):
                drop = self.inputs[0]
                output = self.outputs[0]
                allbytes = droputils.allDropContents(drop)
                buf = io.BytesIO()
                for c in allbytes:
                    if c == b" " or c == b"\n":
                        output.write(buf.getvalue()[::-1])
                        output.write(bytes([c]))
                        buf = io.BytesIO()
                    else:
                        buf.write(bytes([c]))

        a = InMemoryDROP("oid:A", "uid:A")
        b = GrepResult("oid:B", "uid:B", substring=b"a")
        c = InMemoryDROP("oid:C", "uid:C")
        d = SortResult("oid:D", "uid:D")
        e = InMemoryDROP("oid:E", "uid:E")
        f = RevResult("oid:F", "oid:F")
        g = InMemoryDROP("oid:G", "uid:G")

        a.addConsumer(b)
        b.addOutput(c)
        c.addConsumer(d)
        d.addOutput(e)
        e.addConsumer(f)
        f.addOutput(g)

        # Initial write
        contents = b"first line\nwe have an a here\nand another one\nnoone knows me"
        cResExpected = b"we have an a here\nand another one\n"
        eResExpected = b"and another one\nwe have an a here\n"
        gResExpected = b"dna rehtona eno\new evah na a ereh\n"

        with DROPWaiterCtx(self, g):
            a.write(contents)
            a.setCompleted()

        # Get intermediate and final results and compare
        actualRes = []
        for i in [c, e, g]:
            actualRes.append(droputils.allDropContents(i))
        map(
            lambda x, y: self.assertEqual(x, y),
            [cResExpected, eResExpected, gResExpected],
            actualRes,
        )

    def test_errorState(self):
        a = InMemoryDROP("a", "a")
        b = SumupContainerChecksum("b", "b")
        c = InMemoryDROP("c", "c")
        c.addProducer(b)
        b.addInput(a)
        a.setError()
        self.assertEqual(DROPStates.ERROR, a.status)
        self.assertEqual(DROPStates.ERROR, b.status)
        self.assertEqual(DROPStates.ERROR, c.status)

    def test_branch_failure(self):
        self.branch_failure(False)

    def test_branch_too_many_failures(self):
        self.branch_failure(True)

    def branch_failure(self, tooManyFailures):
        """
        Using the container data object to implement a join/barrier dataflow.

        A1, A2 and A3 are FileDROPs
        B1, B2 and B3 are SumupContainerChecksum
        C1, C2 and C3 are InMemoryDROPs
        D is a SumupContainerChecksum
        E is a InMemoryDROP

        --> A1 --> B1 --> C1 --|
        --> A2 --> B2 --> C2 --|--> D --> E
        --> A3 --> B3 --> C3 --|

        Upon writing all A* DROPs, the execution of B* DROPs should be triggered,
        after which "C" will transition to COMPLETE. Once all "C"s have moved to
        COMPLETED "D"'s execution will also be triggered, and finally E will
        hold the sum of B1, B2 and B3's checksums
        """

        # create file data objects
        a1 = InMemoryDROP("oid:A1", "uid:A1")
        a2 = InMemoryDROP("oid:A2", "uid:A2")
        a3 = InMemoryDROP("oid:A3", "uid:A3")

        # CRC Result DROPs, storing the result in memory
        b1 = SumupContainerChecksum("oid:B1", "uid:B1")
        b2 = SumupContainerChecksum("oid:B2", "uid:B2")
        b3 = SumupContainerChecksum("oid:B3", "uid:B3")
        c1 = InMemoryDROP("oid:C1", "uid:C1")
        c2 = InMemoryDROP("oid:C2", "uid:C2")
        c3 = InMemoryDROP("oid:C3", "uid:C3")

        # The final DROP that sums up the CRCs from the container DROP
        d = SumupContainerChecksum("oid:D", "uid:D", input_error_threshold=33)
        e = InMemoryDROP("oid:E", "uid:E")

        # Wire together
        dropAList = [a1, a2, a3]
        dropBList = [b1, b2, b3]
        dropCList = [c1, c2, c3]
        for dropA, dropB in zip(dropAList, dropBList):
            dropA.addConsumer(dropB)
        for dropB, dropC in zip(dropBList, dropCList):
            dropB.addOutput(dropC)
        for dropC in dropCList:
            dropC.addConsumer(d)
        d.addOutput(e)

        # Write data into the initial "A" DROPs, which should trigger
        # the whole chain explained above
        with DROPWaiterCtx(self, e):
            # for dropA in dropAList: # this should be parallel for
            a1.write(b" ")
            a1.setCompleted()
            if tooManyFailures:
                a2.setError()
            else:
                a2.write(b" ")
                a2.setCompleted()
            a3.setError()

        if tooManyFailures:
            completedDrops = dropAList[0:1] + dropBList[0:1] + dropCList[0:1]
            errorDrops = dropAList[1:] + dropBList[1:] + dropCList[1:] + [d, e]
        else:
            completedDrops = dropAList[0:2] + dropBList[0:2] + dropCList[0:2] + [d, e]
            errorDrops = dropAList[2:] + dropBList[2:] + dropCList[2:]

        for drop in completedDrops:
            self.assertEqual(drop.status, DROPStates.COMPLETED)
        for drop in errorDrops:
            self.assertEqual(drop.status, DROPStates.ERROR)

        # The results we want to compare
        # (only in case that at least two branches executed)
        if not tooManyFailures:
            sum_crc = c1.checksum + c2.checksum
            dropEData = int(droputils.allDropContents(e))

            self.assertNotEqual(sum_crc, 0)
            self.assertEqual(sum_crc, dropEData)

    def test_join(self):
        """
        Using the container data object to implement a join/barrier dataflow.

        A1, A2 and A3 are FileDROPs
        B1, B2 and B3 are SumupContainerChecksum
        C1, C2 and C3 are InMemoryDROPs
        D is a SumupContainerChecksum
        E is a InMemoryDROP

        --> A1 --> B1 --> C1 --|
        --> A2 --> B2 --> C2 --|--> D --> E
        --> A3 --> B3 --> C3 --|

        Upon writing all A* DROPs, the execution of B* DROPs should be triggered,
        after which "C" will transition to COMPLETE. Once all "C"s have moved to
        COMPLETED "D"'s execution will also be triggered, and finally E will
        hold the sum of B1, B2 and B3's checksums
        """

        filelen = self._test_drop_sz * ONE_MB
        # create file data objects
        a1 = FileDROP("oid:A1", "uid:A1", expectedSize=filelen)
        a2 = FileDROP("oid:A2", "uid:A2", expectedSize=filelen)
        a3 = FileDROP("oid:A3", "uid:A3", expectedSize=filelen)

        # CRC Result DROPs, storing the result in memory
        b1 = SumupContainerChecksum("oid:B1", "uid:B1")
        b2 = SumupContainerChecksum("oid:B2", "uid:B2")
        b3 = SumupContainerChecksum("oid:B3", "uid:B3")
        c1 = InMemoryDROP("oid:C1", "uid:C1")
        c2 = InMemoryDROP("oid:C2", "uid:C2")
        c3 = InMemoryDROP("oid:C3", "uid:C3")

        # The final DROP that sums up the CRCs from the container DROP
        d = SumupContainerChecksum("oid:D", "uid:D")
        e = InMemoryDROP("oid:E", "uid:E")

        # Wire together
        dropAList = [a1, a2, a3]
        dropBList = [b1, b2, b3]
        dropCList = [c1, c2, c3]
        for dropA, dropB in map(lambda a, b: (a, b), dropAList, dropBList):
            dropA.addConsumer(dropB)
        for dropB, dropC in map(lambda b, c: (b, c), dropBList, dropCList):
            dropB.addOutput(dropC)
        for dropC in dropCList:
            dropC.addConsumer(d)
        d.addOutput(e)

        # Write data into the initial "A" DROPs, which should trigger
        # the whole chain explained above
        with DROPWaiterCtx(self, e):
            for dropA in dropAList:  # this should be parallel for
                for _ in range(self._test_num_blocks):
                    dropA.write(self._test_block)

        # All DROPs are completed now that the chain executed correctly
        for drop in dropAList + dropBList + dropCList:
            self.assertEqual(drop.status, DROPStates.COMPLETED)

        # The results we want to compare
        sum_crc = c1.checksum + c2.checksum + c3.checksum
        dropEData = int(droputils.allDropContents(e))

        self.assertNotEqual(sum_crc, 0)
        self.assertEqual(sum_crc, dropEData)

    def test_app_multiple_outputs(self):
        """
        A small method that tests that the AppDROPs writing to two
        different DROPs outputs works

        The graph constructed by this example looks as follow:

                              |--> E
        A --> B --> C --> D --|
                              |--> F

        Here B and D are an AppDROPs, with D writing to two DROPs
        outputs (E and F) and reading from C. C, in turn, is written by B, which
        in turns reads the data from A
        """

        # This is used as "B"
        class NumberWriterApp(BarrierAppDROP):
            def run(self):
                inputDrop = self.inputs[0]
                output = self.outputs[0]
                howMany = int(droputils.allDropContents(inputDrop))
                for i in range(howMany):
                    output.write(str(i).encode("utf8") + b" ")

        # This is used as "D"
        class OddAndEvenContainerApp(BarrierAppDROP):
            def run(self):
                inputDrop = self.inputs[0]
                outputs = self.outputs

                numbers = droputils.allDropContents(inputDrop).strip().split()
                for n in numbers:
                    outputs[int(n) % 2].write(n + b" ")

        # Create DROPs
        a = InMemoryDROP("oid:A", "uid:A")
        b = NumberWriterApp("oid:B", "uid:B")
        c = InMemoryDROP("oid:A", "uid:A")
        d = OddAndEvenContainerApp("oid:D", "uid:D")
        e = InMemoryDROP("oid:E", "uid:E")
        f = InMemoryDROP("oid:F", "uid:F")

        # Wire them together
        a.addConsumer(b)
        b.addOutput(c)
        c.addConsumer(d)
        d.addOutput(e)
        d.addOutput(f)

        # Start the execution
        with DROPWaiterCtx(self, [e, f]):
            a.write(b"20")
            a.setCompleted()

        # Check the final results are correct
        for drop in [a, b, c, d, e]:
            self.assertEqual(
                drop.status,
                DROPStates.COMPLETED,
                "%r is not yet COMPLETED" % (drop),
            )
        self.assertEqual(
            b"0 2 4 6 8 10 12 14 16 18", droputils.allDropContents(e).strip()
        )
        self.assertEqual(
            b"1 3 5 7 9 11 13 15 17 19", droputils.allDropContents(f).strip()
        )

    def test_dropWroteFromOutside(self):
        """
        A different scenario to those tested above, in which the data
        represented by the DROP isn't actually written *through* the
        DROP. Still, the DROP needs to be moved to COMPLETED once
        the data is written, and reading from it should still yield a correct
        result
        """

        # Write, but not through the DROP
        a = FileDROP("A", "A")
        filename = a.path
        msg = b"a message"
        with open(filename, "wb") as f:
            f.write(msg)
        a.setCompleted()

        # Read from the DROP
        self.assertEqual(msg, droputils.allDropContents(a))
        self.assertIsNotNone(a.checksum)
        self.assertEqual(9, a.size)

        # The drop now calculates the size thus we can't set it anymore
        self.assertRaises(Exception, a.size, len(msg))

    def test_stateMachine(self):
        """
        A simple test to check that some transitions are invalid
        """

        # Nice and easy
        drop = InMemoryDROP("a", "a")
        self.assertEqual(drop.status, DROPStates.INITIALIZED)
        drop.write(b"a")
        self.assertEqual(drop.status, DROPStates.WRITING)
        drop.setCompleted()
        self.assertEqual(drop.status, DROPStates.COMPLETED)

        # Try to overwrite the DROP's checksum and size
        self.assertRaises(Exception, lambda: setattr(drop, "checksum", 0))
        self.assertRaises(Exception, lambda: setattr(drop, "size", 0))

        # Try to write on a DROP that is already COMPLETED
        self.assertRaises(Exception, drop.write, "")

        # Invalid reading on a DROP that isn't COMPLETED yet
        drop = InMemoryDROP("a", "a")
        self.assertRaises(Exception, drop.open)
        self.assertRaises(Exception, drop.read, 1)
        self.assertRaises(Exception, drop.close, 1)

        # Invalid file descriptors used to read/close
        drop.setCompleted()
        fd = drop.open()
        otherFd = random.SystemRandom().randint(0, 1000)
        self.assertNotEqual(fd, otherFd)
        self.assertRaises(Exception, drop.read, otherFd)
        self.assertRaises(Exception, drop.close, otherFd)
        # but using the correct one should be OK
        drop.read(fd)
        self.assertTrue(drop.isBeingRead())
        drop.close(fd)

        # Expire it, then try to set it as COMPLETED again
        drop.status = DROPStates.EXPIRED
        self.assertRaises(Exception, drop.setCompleted)

    def test_externalGraphExecutionDriver(self):
        self._test_graphExecutionDriver(ExecutionMode.EXTERNAL)

    def test_DropGraphExecutionDriver(self):
        self._test_graphExecutionDriver(ExecutionMode.DROP)

    def _test_graphExecutionDriver(self, mode):
        """
        A small test to check that DROPs executions can be driven externally if
        required, and not always internally by themselves
        """
        a = InMemoryDROP("a", "a", executionMode=mode, expectedSize=1)
        b = SumupContainerChecksum("b", "b")
        c = InMemoryDROP("c", "c")
        a.addConsumer(b)
        c.addProducer(b)

        # Write and check
        dropsToWaitFor = [] if mode == ExecutionMode.EXTERNAL else [c]
        with DROPWaiterCtx(self, dropsToWaitFor):
            a.write(b"1")

        if mode == ExecutionMode.EXTERNAL:
            # b hasn't been triggered
            self.assertEqual(c.status, DROPStates.INITIALIZED)
            self.assertEqual(b.status, DROPStates.INITIALIZED)
            self.assertEqual(b.execStatus, AppDROPStates.NOT_RUN)
            # Now let b consume a
            with DROPWaiterCtx(self, [c]):
                b.dropCompleted("a", DROPStates.COMPLETED)
            self.assertEqual(c.status, DROPStates.COMPLETED)
        elif mode == ExecutionMode.DROP:
            # b is already done
            self.assertEqual(c.status, DROPStates.COMPLETED)

    def test_objectAsNormalAndStreamingInput(self):
        """
        A test that checks that a DROP can act as normal and streaming input of
        different AppDROPs at the same time. We use the following graph:

        A --|--> B --> D
            |--> C --> E

        Here B uses A as a streaming input, while C uses it as a normal
        input
        """

        class LastCharWriterApp(AppDROP):
            def initialize(self, **kwargs):
                super(LastCharWriterApp, self).initialize(**kwargs)
                self._lastByte = None

            def dataWritten(self, uid, data):
                self.execStatus = AppDROPStates.RUNNING
                outputDrop = self.outputs[0]
                self._lastByte = data[-1:]
                outputDrop.write(self._lastByte)

            def dropCompleted(self, uid, status):
                self.execStatus = AppDROPStates.FINISHED
                self._notifyAppIsFinished()

        a = InMemoryDROP("a", "a")
        b = LastCharWriterApp("b", "b")
        c = SumupContainerChecksum("c", "c")
        d = InMemoryDROP("d", "d")
        e = InMemoryDROP("e", "e")
        a.addStreamingConsumer(b)
        a.addConsumer(c)
        b.addOutput(d)
        c.addOutput(e)

        # Consumer cannot be normal and streaming at the same time
        self.assertRaises(Exception, a.addConsumer, b)
        self.assertRaises(Exception, a.addStreamingConsumer, c)

        # Write a little, then check the consumers
        def checkDropStates(aStatus, dStatus, eStatus, lastByte):
            self.assertEqual(aStatus, a.status)
            self.assertEqual(dStatus, d.status)
            self.assertEqual(eStatus, e.status)
            if lastByte is not None:
                self.assertEqual(lastByte, b._lastByte)

        checkDropStates(
            DROPStates.INITIALIZED,
            DROPStates.INITIALIZED,
            DROPStates.INITIALIZED,
            None,
        )
        a.write(b"abcde")
        checkDropStates(
            DROPStates.WRITING,
            DROPStates.WRITING,
            DROPStates.INITIALIZED,
            b"e",
        )
        a.write(b"fghij")
        checkDropStates(
            DROPStates.WRITING,
            DROPStates.WRITING,
            DROPStates.INITIALIZED,
            b"j",
        )
        a.write(b"k")
        with DROPWaiterCtx(self, [d, e]):
            a.setCompleted()
        checkDropStates(
            DROPStates.COMPLETED,
            DROPStates.COMPLETED,
            DROPStates.COMPLETED,
            b"k",
        )

        self.assertEqual(b"ejk", droputils.allDropContents(d))

    def test_fileDROP_delete_parent_dir(self):
        """
        A test to check that FileDROPs delete their parent directory upon
        drop.delete() if they are instructed to do so.
        """

        def assertFiles(delete_parent_directory, parentDirExists, tempDir=None):
            tempDir = tempDir or tempfile.mkdtemp()
            a = FileDROP(
                "a",
                "a",
                filepath=tempDir + "/",
                delete_parent_directory=delete_parent_directory,
            )
            a.write(b" ")
            a.setCompleted()
            self.assertTrue(a.exists())
            self.assertTrue(os.path.isdir(tempDir))
            a.delete()
            self.assertFalse(a.exists())
            self.assertEqual(parentDirExists, os.path.isdir(tempDir))
            if parentDirExists:
                shutil.rmtree(tempDir)

        # Test 1: no deletion commanded, directory exists after .delete()
        assertFiles(False, True)
        # Test 2: deletion commanded, directory doesn't exist after .delete()
        assertFiles(True, False)
        # Test 3: deletion commanded, directory not empty, delete still works
        tempDir = tempfile.mkdtemp()
        with open(os.path.join(tempDir, "b"), "wb") as f:
            f.write(b" ")
        assertFiles(True, True, tempDir=tempDir)

    def test_DirectoryDROP(self):
        """
        A small, simple test for the DirectoryDROP
        """

        # Prepare our playground
        tmpdir = tempfile.mkdtemp()
        dirname = f"{tmpdir}/.hidden"
        dirname2 = f"{tmpdir}/.hidden/inside"
        if not os.path.exists(dirname2):
            os.makedirs(dirname2)

        cont1 = DirectoryDROP("e", "e", dirname=dirname)
        cont2 = DirectoryDROP("f", "f", dirname=dirname2)

        print(cont1.path)
        print(cont2.path)


        # Paths are absolutely reported
        self.assertEqual(
            os.path.realpath(f"{tmpdir}/.hidden"), os.path.realpath(cont1.path)
        )
        self.assertEqual(
            os.path.realpath(f"{tmpdir}/.hidden/inside"),
            os.path.realpath(cont2.path),
        )

        self.assertTrue(cont1.exists())
        self.assertTrue(cont2.exists())

        # Revert to previous state
        shutil.rmtree(dirname, True)

    def test_multipleProducers(self):
        """
        A test that checks that multiple-producers correctly drive the state of
        their shared output
        """

        class App(BarrierAppDROP):
            def run(self):
                pass

        a, b, c, d, e = [App(chr(ord("A") + i), chr(ord("A") + i)) for i in range(5)]
        f = InMemoryDROP("F", "F")
        for drop in a, b, c, d, e:
            drop.addOutput(f)

        self.assertEqual(DROPStates.INITIALIZED, f.status)
        for drop in a, b, c, d, e:
            self.assertEqual(AppDROPStates.NOT_RUN, drop.execStatus)

        # Run the first 4 ones, F should still be in INITIALIZED
        for drop in a, b, c, d:
            drop.execute()
        self.assertEqual(DROPStates.INITIALIZED, f.status)
        self.assertEqual(AppDROPStates.NOT_RUN, e.execStatus)
        for drop in a, b, c, d:
            self.assertEqual(AppDROPStates.FINISHED, drop.execStatus)

        # Run the final one, now F should be COMPLETED
        e.execute()
        self.assertEqual(DROPStates.COMPLETED, f.status)
        for drop in a, b, c, d, e:
            self.assertEqual(AppDROPStates.FINISHED, drop.execStatus)

    def test_eager_inputFired_app(self):
        """
        Tests that InputFiredApps works as expected
        """

        class App(BarrierAppDROP):
            def run(self):
                pass

        # No n_effective_inputs given
        cause = InvalidDropException
        run_errormanagement_exception_test(self, InputFiredAppDROP,
                                           cause, "a", "a")
        # Invalid values
        run_errormanagement_exception_test(
            self, InputFiredAppDROP,
            cause,
            "a", "a", n_effective_inputs=-2
        )

        run_errormanagement_exception_test(
            self, InputFiredAppDROP,
            cause,
            "a", "a", n_effective_inputs=0
        )

        # More effective inputs than inputs
        a = InMemoryDROP("b", "b")
        b = InputFiredAppDROP("a", "a", n_effective_inputs=2)
        b.addInput(a)
        self.assertRaises(RuntimeError, a.setCompleted)

        # 2 effective inputs, 4 outputs. Trigger 2 inputs and make sure the
        # app has run
        a, b, c, d = [InMemoryDROP(str(i), str(i)) for i in range(4)]
        e = App("e", "e", n_effective_inputs=2)
        for x in a, b, c, d:
            e.addInput(x)

        with DROPWaiterCtx(self, e, 5):
            a.setCompleted()
            b.setCompleted()

        self.assertEqual(AppDROPStates.FINISHED, e.execStatus)
        self.assertEqual(DROPStates.COMPLETED, a.status)
        self.assertEqual(DROPStates.COMPLETED, b.status)
        self.assertEqual(DROPStates.INITIALIZED, c.status)
        self.assertEqual(DROPStates.INITIALIZED, d.status)

    def test_n_tries_app(self):
        class FailOnlyTheFirstTimeApp(BarrierAppDROP):
            def initialize(self, **kwargs):
                BarrierAppDROP.initialize(self, **kwargs)
                self.i = 0

            def run(self):
                if self.i == 0:
                    self.i = 1
                    raise Exception

        # Check that we have a normal failure with the default values
        a = FailOnlyTheFirstTimeApp("a", "a")
        a.execute()
        self.assertEqual(DROPStates.ERROR, a.status)
        self.assertEqual(AppDROPStates.ERROR, a.execStatus)

        # But it should run if we specify a bigger amount of tries
        a = FailOnlyTheFirstTimeApp("a", "a", n_tries=2)
        a.execute()
        self.assertEqual(DROPStates.COMPLETED, a.status)
        self.assertEqual(AppDROPStates.FINISHED, a.execStatus)

    def test_rdbms_drop(self):
        dbfile = f"{tempfile.mkdtemp()}/test_rdbms_drop.db"
        if os.path.isfile(dbfile):
            os.unlink(dbfile)

        with contextlib.closing(sqlite3.connect(dbfile)) as conn:  # @UndefinedVariable
            with contextlib.closing(conn.cursor()) as cur:
                cur.execute(
                    "CREATE TABLE super_mega_table(a_string varchar(64) PRIMARY KEY, an_integer integer)"
                )

        try:
            a = RDBMSDrop(
                "a",
                "a",
                dbmodule="sqlite3",
                dbtable="super_mega_table",
                dbparams={"database": dbfile},
            )
            a.insert({"a_string": "hello", "an_integer": 0})
            a.insert({"a_string": "hello1", "an_integer": 1})

            res = a.select(columns=("an_integer",))
            self.assertEqual(2, len(res))

            res = a.select(columns=("an_integer",), condition="an_integer < 1")
            self.assertEqual(1, len(res))
            self.assertEqual(0, res[0][0])
        finally:
            os.unlink(dbfile)


class TestDROPReproducibility(unittest.TestCase):
    def test_drop_rerun(self):
        a = NullDROP("a", "a")
        a.reproducibility_level = ReproducibilityFlags.RERUN
        a.setCompleted()
        self.assertEqual(a.generate_merkle_data(), {"status": DROPStates.COMPLETED})
        self.assertIsNotNone(a.merkleroot)

    def test_drop_repeat(self):
        a = NullDROP("a", "a")
        a.reproducibility_level = ReproducibilityFlags.REPEAT
        a.setCompleted()
        self.assertEqual(a.generate_merkle_data(), {"status": DROPStates.COMPLETED})
        self.assertIsNotNone(a.merkleroot)
        pass

    def test_drop_recompute(self):
        a = NullDROP("a", "a")
        a.reproducibility_level = ReproducibilityFlags.RECOMPUTE
        a.setCompleted()
        self.assertEqual(a.generate_merkle_data(), {"status": DROPStates.COMPLETED})
        self.assertIsNotNone(a.merkleroot)
        pass

    def test_drop_reproduce(self):
        a = NullDROP("a", "a")
        a.reproducibility_level = ReproducibilityFlags.REPRODUCE
        a.setCompleted()
        self.assertEqual(a.generate_merkle_data(), {})
        self.assertIsNone(a.merkleroot)
        pass

    def test_drop_replicate_sci(self):
        a = NullDROP("a", "a")
        a.reproducibility_level = ReproducibilityFlags.REPLICATE_SCI
        a.setCompleted()
        self.assertEqual(a.generate_rerun_data(), {"status": DROPStates.COMPLETED})
        self.assertIsNotNone(a.merkleroot)
        pass

    def test_drop_replicate_comp(self):
        a = NullDROP("a", "a")
        a.reproducibility_level = ReproducibilityFlags.REPLICATE_COMP
        a.setCompleted()
        self.assertEqual(a.generate_rerun_data(), {"status": DROPStates.COMPLETED})
        self.assertIsNotNone(a.merkleroot)
        pass

    def test_drop_replicate_total(self):
        a = NullDROP("a", "a")
        a.reproducibility_level = ReproducibilityFlags.REPLICATE_TOTAL
        a.setCompleted()
        self.assertEqual(a.generate_rerun_data(), {"status": DROPStates.COMPLETED})
        self.assertIsNotNone(a.merkleroot)
        pass

    def test_file_reproducibility(self):
        from dlg.common.reproducibility.reproducibility import common_hash

        data = b"Helloworld"
        data_hash = common_hash(data)
        a = FileDROP("a", "a")
        a.write(data)
        a.reproducibility_level = ReproducibilityFlags.RERUN
        a.setCompleted()
        b = NullDROP("b", "b")
        b.reproducibility_level = ReproducibilityFlags.RERUN
        b.setCompleted()
        self.assertEqual(a.merkleroot, b.merkleroot)

        a.reproducibility_level = ReproducibilityFlags.REPEAT
        a.commit()
        self.assertEqual(a.merkleroot, b.merkleroot)

        a.reproducibility_level = ReproducibilityFlags.RECOMPUTE
        a.commit()
        self.assertEqual(a.merkleroot, b.merkleroot)

        a.reproducibility_level = ReproducibilityFlags.REPRODUCE
        a.commit()
        self.assertNotEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(a.generate_merkle_data(), {"data_hash": data_hash})

        a.reproducibility_level = ReproducibilityFlags.REPLICATE_SCI
        a.commit()
        self.assertNotEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(
            a.generate_merkle_data(),
            {"data_hash": data_hash, "status": DROPStates.COMPLETED},
        )

        a.reproducibility_level = ReproducibilityFlags.REPLICATE_COMP
        a.commit()
        self.assertNotEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(
            a.generate_merkle_data(),
            {"data_hash": data_hash, "status": DROPStates.COMPLETED},
        )

        a.reproducibility_level = ReproducibilityFlags.REPLICATE_TOTAL
        a.commit()
        self.assertNotEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(
            a.generate_merkle_data(),
            {"data_hash": data_hash, "status": DROPStates.COMPLETED},
        )

    def test_memory_reproducibility(self):
        from dlg.common.reproducibility.reproducibility import common_hash

        data = b"Helloworld"
        data_hash = common_hash(data)
        a = InMemoryDROP("a", "a")
        a.write(data)
        a.reproducibility_level = ReproducibilityFlags.RERUN
        a.setCompleted()
        b = NullDROP("b", "b")
        b.reproducibility_level = ReproducibilityFlags.RERUN
        b.setCompleted()
        self.assertEqual(a.merkleroot, b.merkleroot)

        a.reproducibility_level = ReproducibilityFlags.REPEAT
        a.commit()
        self.assertEqual(a.merkleroot, b.merkleroot)

        a.reproducibility_level = ReproducibilityFlags.RECOMPUTE
        a.commit()
        self.assertEqual(a.merkleroot, b.merkleroot)

        a.reproducibility_level = ReproducibilityFlags.REPRODUCE
        a.commit()
        self.assertNotEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(a.generate_merkle_data(), {"data_hash": data_hash})

        a.reproducibility_level = ReproducibilityFlags.REPLICATE_SCI
        a.commit()
        self.assertNotEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(
            a.generate_merkle_data(),
            {"data_hash": data_hash, "status": DROPStates.COMPLETED},
        )

        a.reproducibility_level = ReproducibilityFlags.REPLICATE_COMP
        a.commit()
        self.assertNotEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(
            a.generate_merkle_data(),
            {"data_hash": data_hash, "status": DROPStates.COMPLETED},
        )

        a.reproducibility_level = ReproducibilityFlags.REPLICATE_TOTAL
        a.commit()
        self.assertNotEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(
            a.generate_merkle_data(),
            {"data_hash": data_hash, "status": DROPStates.COMPLETED},
        )

    def test_rdbms_reproducibility(self):

        dbfile = f"{tempfile.mkdtemp()}/test_rdbms_drop.db"
        if os.path.isfile(dbfile):
            os.unlink(dbfile)

        b = NullDROP("b", "b")
        b.reproducibility_level = ReproducibilityFlags.RERUN
        b.setCompleted()

        with contextlib.closing(sqlite3.connect(dbfile)) as conn:  # @UndefinedVariable
            with contextlib.closing(conn.cursor()) as cur:
                cur.execute(
                    "CREATE TABLE super_mega_table(a_string varchar(64) PRIMARY KEY, an_integer integer)"
                )

        try:
            a = RDBMSDrop(
                "a",
                "a",
                dbmodule="sqlite3",
                dbtable="super_mega_table",
                dbparams={"database": dbfile},
            )
            a.insert({"a_string": "hello", "an_integer": 0})
            a.insert({"a_string": "hello1", "an_integer": 1})
            a.reproducibility_level = ReproducibilityFlags.RERUN
            a.setCompleted()
            self.assertEqual(a.merkleroot, b.merkleroot)

            a.reproducibility_level = ReproducibilityFlags.REPEAT
            a.commit()
            self.assertEqual(a.merkleroot, b.merkleroot)

            a.reproducibility_level = ReproducibilityFlags.RECOMPUTE
            a.commit()
            self.assertEqual(a.merkleroot, b.merkleroot)

            a.reproducibility_level = ReproducibilityFlags.REPRODUCE
            a.commit()
            self.assertEqual(a.generate_merkle_data(), {"query_log": a._querylog})
            self.assertNotEqual(a.merkleroot, b.merkleroot)

            a.reproducibility_level = ReproducibilityFlags.REPLICATE_SCI
            a.commit()
            self.assertNotEqual(a.merkleroot, b.merkleroot)
            self.assertEqual(
                a.generate_merkle_data(),
                {"query_log": a._querylog, "status": DROPStates.COMPLETED},
            )

            a.reproducibility_level = ReproducibilityFlags.REPLICATE_COMP
            a.commit()
            self.assertNotEqual(a.merkleroot, b.merkleroot)
            self.assertEqual(
                a.generate_merkle_data(),
                {"query_log": a._querylog, "status": DROPStates.COMPLETED},
            )

            a.reproducibility_level = ReproducibilityFlags.REPLICATE_TOTAL
            a.commit()
            self.assertNotEqual(a.merkleroot, b.merkleroot)
            self.assertEqual(
                a.generate_merkle_data(),
                {"query_log": a._querylog, "status": DROPStates.COMPLETED},
            )
        finally:
            os.unlink(dbfile)


def func1(result:bool=False):
    return result


class BranchAppDropTestsBase(object):
    """Tests for the Branch class"""

    def _simple_branch_with_outputs(self, result, uids):
        a = Branch(uids[0], uids[0], result=result, func_name="test.test_drop.func1")
        b, c = (self.DataDropType(x, x) for x in uids[1:])
        a.addOutput(b)
        a.addOutput(c)
        a.true = c
        a.false = b
        return a, b, c

    def _assert_drop_in_status(self, drop, status, execStatus):
        self.assertEqual(
            drop.status, status, f"{drop} has status {drop.status} != {status}"
        )
        if isinstance(drop, AppDROP):
            self.assertEqual(
                drop.execStatus,
                execStatus,
                f"{drop} has execStatus {drop.execStatus} != {execStatus}",
            )

    def _assert_drop_complete_or_skipped(self, drop, complete_expected):
        if complete_expected:
            self._assert_drop_in_status(
                drop, DROPStates.COMPLETED, AppDROPStates.FINISHED
            )
        else:
            self._assert_drop_in_status(drop, DROPStates.SKIPPED, AppDROPStates.SKIPPED)

    def _test_single_branch_graph(self, result, levels):
        """
        Test this graph, where each level appends one drop to each branch

            ,-- true --> B --> ...
        A ---- false --> C --> ...
        """
        a, b, c = self._simple_branch_with_outputs(result, "abc")
        # This order is important, since we are using indexed ports in Branch
        last_false = b
        last_true = c
        all_drops = [a, b, c]

        # all_uids is ['de', 'fg', 'hi', ....]
        all_uids = [
            string.ascii_lowercase[i : i + 2]
            for i in range(3, len(string.ascii_lowercase), 2)
        ]

        for level, uids in zip(range(levels), all_uids):
            if level % 2:
                x, y = (self.DataDropType(uid, uid) for uid in uids)
                last_true.addOutput(x)
                last_false.addOutput(y)
            else:
                x, y = (NullBarrierApp(uid, uid) for uid in uids)
                last_true.addConsumer(x)
                last_false.addConsumer(y)
            all_drops += [x, y]
            last_true = x
            last_false = y

        with DROPWaiterCtx(
            self,
            [last_true, last_false],
            2,
            [DROPStates.COMPLETED, DROPStates.SKIPPED],
        ):
            a.async_execute()
        time.sleep(0.01)
        # Depending on "result", the "true" branch will be run or skipped
        self._assert_drop_complete_or_skipped(last_true, result)
        self._assert_drop_complete_or_skipped(last_false, not result)

    def test_Branch_true_first(self):
        """
        Test condition met, true branch connected first.

        """
        value = 2
        b = Branch("b", "b", func_name="condition", func_code="def condition(x): return x>0", x=value)
        t = self.DataDropType("t", "t", type="int")
        f = self.DataDropType("f", "f", type="int")
        b.addOutput(t)
        b.false = f
        b.true = t
        b.addOutput(f)
        b.execute()
        res = pickle.loads(droputils.allDropContents(t))
        self.assertEqual(value, res)

    def test_Branch_false_first(self):
        """
        Test condition met, false branch connected first.

        """
        value = 2
        b = Branch("b", "b", func_name="condition", func_code="def condition(x): return x>0", x=value)
        f = self.DataDropType("f", "f", type="int")
        t = self.DataDropType("t", "t", type="int")
        b.addOutput(t)
        b.false = f
        b.true = t
        b.addOutput(f)
        b.execute()
        res = pickle.loads(droputils.allDropContents(t))
        self.assertEqual(value, res)

    def test_Branch_false(self):
        """
        Test condition not met.
        """
        value = 1
        b = Branch("b", "b", func_name="condition", func_code="def condition(x): return x>0", x=value)
        f = self.DataDropType("f", "f", type="Integer")
        t = self.DataDropType("t", "t", type="Integer")
        b.addOutput(t)
        b.false = f
        b.true = t
        b.addOutput(f)
        b.execute()
        res = pickle.loads(droputils.allDropContents(t))
        self.assertEqual(value, res)        



    def test_simple_branch(self):
        """Check that simple branch event transmission works"""
        self._test_single_branch_graph(True, 0)
        self._test_single_branch_graph(False, 0)

    def test_simple_branch_one_level(self):
        """Like test_simple_branch_app, but events propagate downstream one level"""
        self._test_single_branch_graph(True, 1)
        self._test_single_branch_graph(False, 1)

    def test_simple_branch_two_levels(self):
        """Like test_skipped_propagates, but events propagate more levels"""
        self._test_single_branch_graph(True, 2)
        self._test_single_branch_graph(False, 2)

    def test_simple_branch_more_levels(self):
        for levels in (3, 4, 5, 6, 7):
            self._test_single_branch_graph(True, levels)
            self._test_single_branch_graph(False, levels)

    def _test_multi_branch_graph(self, *results):
        """
        Test this graph, each level appends a new branch to the first output
        of the previous branch.

            ,- false --> C         ,- false --> F
        A ----- true --> B --> D ----- true --> E --> ...
        """

        a, b, c = self._simple_branch_with_outputs(results[0], "abc")
        all_drops = [a, b, c]
        last_first_output = b

        # all_uids is ['de', 'fg', 'hi', ....]
        all_uids = [
            string.ascii_lowercase[i : i + 3]
            for i in range(4, len(string.ascii_lowercase), 3)
        ]

        for uids, result in zip(all_uids, results[1:]):
            x, y, z = self._simple_branch_with_outputs(result, uids)
            all_drops += [x, y, z]
            last_first_output.addConsumer(x)
            last_first_output = y

        with DROPWaiterCtx(
            self, all_drops, 3, [DROPStates.COMPLETED, DROPStates.SKIPPED]
        ):
            a.async_execute()

        # TODO: Checking each individual drop depending on "results" would be a
        # tricky business, so we are skipping that check for now.

    def test_multi_branch_one_level(self):
        """Check that simple branch event transmission wroks"""
        self._test_multi_branch_graph(True)
        self._test_multi_branch_graph(False)

    def test_multi_branch_two_levels(self):
        """Like test_simple_branch_app, but events propagate downstream one level"""
        # self._test_multi_branch_graph(True, True)
        # self._test_multi_branch_graph(True, False)
        self._test_multi_branch_graph(False, False)

    def test_multi_branch_more_levels(self):
        """Like test_skipped_propagates, but events propagate more levels"""
        for levels in (3, 4, 5, 6, 7):
            self._test_multi_branch_graph(True, levels)
            self._test_multi_branch_graph(False, levels)


class BranchAppDropTestsWithMemoryDrop(BranchAppDropTestsBase, unittest.TestCase):
    DataDropType = InMemoryDROP


class BranchAppDropTestsWithFileDrop(BranchAppDropTestsBase, unittest.TestCase):
    DataDropType = FileDROP


if __name__ == "__main__":
    unittest.main()



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2024
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

"""
This module contains classes and helper-methods to support the various manager classes
"""

import logging
import dlg.constants as constants
from enum import IntEnum

logger = logging.getLogger(f"dlg.{__name__}")


class NodeProtocolPosition(IntEnum):
    HOST = 0
    PORT = 1
    EVENTS_PORT = 2
    RPC_PORT = 3


class Node:
    """
    Class for encapsulating compute node information to standardise
    inter-node communication.
    """

    def __init__(self, host: str, dim=False):
        try:
            chunks = host.split(":")
            num_chunks = len(chunks)
            self.host = constants.NODE_DEFAULT_HOSTNAME
            self.port = (
                constants.ISLAND_DEFAULT_REST_PORT
                if dim
                else constants.NODE_DEFAULT_REST_PORT
            )
            self.events_port = constants.NODE_DEFAULT_EVENTS_PORT
            self.rpc_port = constants.NODE_DEFAULT_RPC_PORT
            self._rest_port_specified = False

            if num_chunks >= 1:
                self.host = chunks[NodeProtocolPosition.HOST]
            if num_chunks >= 2:
                self.port = self._validate_port(chunks[NodeProtocolPosition.PORT])
                self._rest_port_specified = True
            if num_chunks >= 3:
                self.events_port = self._validate_port(
                    chunks[NodeProtocolPosition.EVENTS_PORT]
                )
            if num_chunks >= 4:
                self.rpc_port = self._validate_port(
                    chunks[NodeProtocolPosition.RPC_PORT]
                )
        except AttributeError as e:
            raise RuntimeError(
                f"Constructor has been passed non-string object and cannot"
                f"be converted to Node: type {type(host)}.",
            ) from e
        except ValueError as e:
            logger.error(
                "An issue has occured with translating node information", exc_info=e
            )

    def _validate_port(self, port: str) -> int:
        """
        Confirm the port provided is within the correct range of ports and returns it
        as an integer.

        Param:

        Raises: Invalid port

        :return: int, integer representation of port passed to the command line.
        """
        validated_port = int(port)
        if not 1 <= validated_port <= 65535:
            raise ValueError("Port numbers must be between 1 and 65535")
        else:
            return validated_port

    def serialize(self):
        """
        Convert to the expect string representation of our Node using the
        following 'protocol':

            "host:port:event_port:rpc_port"

        :return: str
        """
        return f"{self.host}:{self.port}:{self.events_port}:{self.rpc_port}"

    @property
    def rest_port_specified(self):
        """
        Returns True if we specified a Node REST port when passing the list of nodes to
        the DIM at startup.
        """
        return self._rest_port_specified

    def __str__(self):
        """
        Make our serialized Node the string.
        :return: str
        """
        return self.serialize()

    def __repr__(self):
        """ """
        return self.serialize()

    def __eq__(self, other):
        if isinstance(other, Node):
            return hash(self) == hash(other)
        if isinstance(other, str):
            return hash(self) == hash(other)

    def __hash__(self):
        return hash(str(self))



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2014
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Module extends the multiprocessing.process class to include some convenient accouterments.
Based on approach in https://stackoverflow.com/questions/19924104/python-multiprocessing-handling-child-errors-in-parent
"""
import multiprocessing
import traceback


class DlgProcess(multiprocessing.Process):
    def __init__(self, *args, **kwargs):
        multiprocessing.Process.__init__(self, *args, **kwargs)
        self._parentConn, self._childConn = multiprocessing.Pipe()
        self._exception = None

    def run(self):
        try:
            multiprocessing.Process.run(self)
            self._childConn.send(None)
        except multiprocessing.ProcessError as e:
            tb = traceback.format_exc()
            self._childConn.send((e, tb))

    @property
    def exception(self):
        if self._parentConn.poll():
            e = self._parentConn.recv()
            if e:
                e = type(e[0])(e[1])
            self._exception = e
        return self._exception



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2025
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

"""
Logging facilities for DALiuGE runtime environment.
"""

import logging

PREFIX = "dlg"

USER = 25

def setup_logger_class():

    # Doc building is failing due with "ValueError: Empty module name" in
    # "from . import drop" further below, which I can't easily explain.
    # Just skip this setup in RTD for the time being
    import os

    if os.environ.get("READTHEDOCS", None) == "True":
        return

    class _DlgLogger(logging.Logger):
        def makeRecord(self, *args, **kwargs):
            record = super(_DlgLogger, self).makeRecord(*args, **kwargs)

            from dlg import drop
            from dlg.manager import session

            # Try to get the UID of the drop ultimately in charge of sending this
            # log record, if any
            try:
                drop = drop.track_current_drop.tlocal.drop
            except AttributeError:
                drop = None
            try:
                drop_uid = drop.humanKey if drop else ""
            except AttributeError:
                drop_uid = ""

            # Do the same with the session_id, which can be found via the drop (if any)
            # or checking if there is a session currently executing something
            session_id = ""
            if drop and hasattr(drop, "dlg_session_id"):
                session_id = drop.dlg_session_id
            else:
                try:
                    session_id = session.track_current_session.tlocal.session.sessionId
                except AttributeError:
                    pass
            record.drop_uid = drop_uid
            record.session_id = session_id
            return record

        def user(self, message, *args, **kwargs):
            """
            Operator-specific error messages

            :param self:
            :param message:
            :param args:
            :param kwargs:
            :return:
            """
            if self.isEnabledFor(USER):
                self._log(USER, message, args, **kwargs)

    logging.addLevelName(USER, "USER")
    # To avoid 'No handlers could be found for logger' messages during testing
    logging.getLogger(__name__).addHandler(logging.NullHandler())

    # Use our own logger class, which knows about the currently executing app
    logging.setLoggerClass(_DlgLogger)
    # logging.logger.user = _Dlglogger.user
    logging.addLevelName(USER, "USER")
    # logging.user = _DlgLogger.user
    logging.USER = USER # CREATE



# class RedirectOperatorLoggingHandler(logging.Handler):
#     def __init__(self, target_logger):
#         super().__init__()
#         self.target_logger = target_logger
#         self.setFormatter(logging.Formatter('%(name)s - %(message)s'))
#
#     def emit(self, record):
#         msg = self.format(record)
#         self.target_logger.trace(msg)  # Send to your TRACE level


#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
from dlg.ddap_protocol import DROPStates
from dlg.data.drops.data_base import DataDROP, logger, track_current_drop
from dlg.data.io import NgasIO, NgasLiteIO
from dlg.meta import dlg_string_param, dlg_int_param


##
# @brief NGAS
# @details An archive on the Next Generation Archive System (NGAS).
# @par EAGLE_START
# @param category NGAS
# @param tag daliuge
# @param ngasSrv localhost/String/ComponentParameter/NoPort/ReadWrite//False/False/The URL of the NGAS Server
# @param ngasPort 7777/Integer/ComponentParameter/NoPort/ReadWrite//False/False/The port of the NGAS Server
# @param ngasFileId /String/ComponentParameter/NoPort/ReadWrite//False/False/File ID on NGAS (for retrieval only)
# @param ngasConnectTimeout 2/Integer/ComponentParameter/NoPort/ReadWrite//False/False/Timeout for connecting to the NGAS server
# @param ngasMime "text/ascii"/String/ComponentParameter/NoPort/ReadWrite//False/False/Mime-type to be used for archiving
# @param ngasTimeout 2/Integer/ComponentParameter/NoPort/ReadWrite//False/False/Timeout for receiving responses for NGAS
# @param block_skip False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/If set the drop will block a skipping chain until the last producer has finished and is not also skipped.
# @param io /Object/ApplicationArgument/InputOutput/ReadWrite//False/False/Input Output port
# @param dropclass dlg.data.drops.ngas.NgasDROP/String/ComponentParameter/NoPort/ReadWrite//False/False/Drop class
# @param base_name ngas/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in this node
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param streaming False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component streams input and output data
# @param persist True/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component contains data that should not be deleted after execution
# @par EAGLE_END
class NgasDROP(DataDROP):
    """
    A DROP that points to data stored in an NGAS server
    """

    ngasSrv = dlg_string_param("ngasSrv", "localhost")
    ngasPort = dlg_int_param("ngasPort", 7777)
    ngasFileId = dlg_string_param("ngasFileId", None)
    ngasTimeout = dlg_int_param("ngasTimeout", 2)
    ngasConnectTimeout = dlg_int_param("ngasConnectTimeout", 2)
    ngasMime = dlg_string_param("ngasMime", "application/octet-stream")
    len = dlg_int_param("len", -1)
    ngas_checksum = None

    def initialize(self, **kwargs):
        if self.len == -1:
            # TODO: For writing the len field should be set to the size of the input drop
            self.len = self._size
        if self.ngasFileId:
            self.fileId = self.ngasFileId
        else:
            self.fileId = self.uid
        # self.path = self.fileId # we are using the path for logging

    def getIO(self):
        try:
            ngasIO = NgasIO(
                self.ngasSrv,
                self.fileId,
                self.ngasPort,
                self.ngasConnectTimeout,
                self.ngasTimeout,
                length=self.len,
                mimeType=self.ngasMime,
            )
        except ImportError:
            logger.warning("NgasIO not available, using NgasLiteIO instead")
            ngasIO = NgasLiteIO(
                self.ngasSrv,
                self.fileId,
                self.ngasPort,
                self.ngasConnectTimeout,
                self.ngasTimeout,
                length=self.len,
                mimeType=self.ngasMime,
            )
        return ngasIO

    @track_current_drop
    def setCompleted(self):
        """
        Override this method in order to get the size of the drop set once it is completed.
        """
        if not self._setCompletedStateCheck():
            return

        self._closeWriters()

        # here we set the size. It could happen that nothing is written into
        # this file, in which case we create an empty file so applications
        # downstream don't fail to read
        logger.debug("Trying to set size of NGASDrop")
        if not self.getIO().exists():
            self.setError()
            raise FileNotFoundError(f"File with fileId '{self.fileId}' not found on host '{self.ngasSrv}'!")
        try:
            stat = self.getIO().fileStatus()
            logger.debug(
                "Setting size of NGASDrop %s to %s",
                self.fileId,
                stat["FileSize"],
            )
            self._size = int(stat["FileSize"])
            self.ngas_checksum = str(stat["Checksum"])
        except:
            # we''ll try this again in case there is some other issue
            # try:
            #     with open(self.path, 'wb'):
            #         pass
            # except:
            #     self.status = DROPStates.ERROR
            #     logger.error("Path not accessible: %s" % self.path)
            logger.debug("Setting size of NGASDrop to %s", 0)
            self._size = 0
            self.setError()
            raise
        # Signal our subscribers that the show is over
        logger.debug("Moving %r to COMPLETED", self)
        self.status = DROPStates.COMPLETED
        self._fire("dropCompleted", status=DROPStates.COMPLETED)
        self.completedrop()

    @property
    def dataURL(self) -> str:
        return "ngas://%s:%s/%s" % (self.ngasSrv, self.ngasPort, self.fileId)

    # Override
    def generate_reproduce_data(self):
        if self.ngas_checksum is None or self.ngas_checksum == "":
            return {"fileid": self.ngasFileId, "size": self._size}
        return {"data_hash": self.ngas_checksum}



from concurrent.futures import Future
from dlg.apps.app_base import InputFiredAppDROP
from dlg.event import Event, EventHandler
import pytest


class MockThrowingDrop(InputFiredAppDROP):
    def run(self):
        raise RuntimeError("Drop throw")


class MockThrowingHandler(EventHandler):
    def handleEvent(self, e: Event) -> None:
        raise RuntimeError("Handler throw")


def test_async_execute_catches_and_logs_unexpected_exception(
    caplog: pytest.LogCaptureFixture,
):
    drop = MockThrowingDrop("t", "t", n_effective_inputs=1)
    handler = MockThrowingHandler()
    drop.subscribe(handler)

    fut = drop.async_execute()
    assert isinstance(fut, Future)
    fut.result()

    assert "Handler throw" in caplog.text
    # execute should handle exceptions in the run method
    assert "Drop throw" not in caplog.text


def test_execute_propogates_unexpected_exception():
    drop = MockThrowingDrop("t", "t", n_effective_inputs=1)
    handler = MockThrowingHandler()
    drop.subscribe(handler)

    with pytest.raises(RuntimeError) as e:
        drop.execute()

    assert "Handler throw" in str(e.value)
    assert "Drop throw" not in str(e.value)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2019
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import collections
import logging
import os
import pickle
import time
import re
import socket

from . import deployment_utils

logger = logging.getLogger(f"dlg.{__name__}")


class Remote(object):
    def __init__(self, options, my_ip):
        self.options = options
        self.my_ip = my_ip
        self.num_islands = options.num_islands
        self.run_proxy = options.monitor_host
        self.co_host_dim = options.co_host_dim
        self.rank = None
        self.size = None
        self.sorted_peers = None

    def _get_ip_from_name(self, hostname):
        rx = re.compile(r"^(?:[0-9]{1,3}\.){3}[0-9]{1,3}$")
        if rx.match(hostname):  # already an IP??
            return hostname
        else:
            return socket.gethostbyname(hostname)

    def _set_world(self, rank, size, sorted_peers):
        """Called by subclasses"""
        self.rank = rank
        self.size = size
        if len(set(sorted_peers)) != self.size:
            raise RuntimeError("More than one task started per node, cannot continue")
        # convert nodes to IP addresses if hostnames
        self.sorted_peers = list(map(self._get_ip_from_name, sorted_peers))
        nm_range = self._nm_range()
        if nm_range[0] == nm_range[1]:
            raise RuntimeError(
                "No nodes left for Node Managers with the requested setup"
            )

    def _dim_range(self):
        first = 0
        if self.num_islands > 1:
            first = 1
        if self.run_proxy:
            first += 1
        return first, first + self.num_islands

    def _nm_range(self):
        if self.co_host_dim:
            first = 0
        else:
            first = 1
        if self.run_proxy:
            first = 2
        if self.num_islands > 1:
            first += self.num_islands
        return first, self.size

    def _dim_nms(self, pg):
        dim_nodes = collections.defaultdict(set)
        ranks = {ip: rank for rank, ip in enumerate(self.sorted_peers)}
        for drop in pg:
            nm, dim = drop["node"], drop["island"]
            dim_nodes[dim].add(nm)
        for dim in self.dim_ips:
            r = ranks[dim]
            nms = list(dim_nodes[dim])
            logger.debug("Nodes for DIM %s (rank %d): %r", dim, r, nms)
            yield r, nms

    @property
    def is_highest_level_manager(self):
        return self.rank == 0

    @property
    def hl_mgr_ip(self):
        return self.sorted_peers[0]

    @property
    def is_dim(self):
        return self.rank in range(*self._dim_range())

    @property
    def dim_ips(self):
        start, end = self._dim_range()
        return self.sorted_peers[start:end]

    @property
    def is_nm(self):
        return self.rank in range(*self._nm_range())

    @property
    def nm_ips(self):
        start, end = self._nm_range()
        return self.sorted_peers[start:end]

    @property
    def is_proxy(self):
        return self.run_proxy and self.rank == 1

    @property
    def proxy_ip(self):
        if not self.run_proxy:
            raise RuntimeError("proxy_ip requested when not running with proxy enabled")
        return self.sorted_peers[1]


class MPIRemote(Remote):
    def __init__(self, options, my_ip):
        super(MPIRemote, self).__init__(options, my_ip)
        from mpi4py import MPI  # @UnresolvedImport

        self.comm = MPI.COMM_WORLD  # @UndefinedVariable
        self._set_world(
            self.comm.Get_rank(), self.comm.Get_size(), self.comm.allgather(self.my_ip)
        )

    def send_dim_nodes(self, pg):
        for dim_rank, nms in self._dim_nms(pg):
            self.comm.send(nms, dest=dim_rank)

    def recv_dim_nodes(self):
        return self.comm.recv(source=0)


class FilesystemBasedRemote(Remote):
    def send_dim_nodes(self, pg):
        basedir = self.options.log_dir
        for dim_rank, nms in self._dim_nms(pg):
            fname = "dim_%d_nodes.pickle" % dim_rank
            stage1_fname = os.path.join(basedir, ".%s" % fname)
            stage2_fname = os.path.join(basedir, "%s" % fname)
            with open(stage1_fname, "wb") as f:
                pickle.dump(nms, f)
            os.rename(stage1_fname, stage2_fname)

    def recv_dim_nodes(self):
        fname = os.path.join(self.options.log_dir, "dim_%d_nodes.pickle" % self.rank)
        while not os.path.exists(fname):
            time.sleep(1)
        with open(fname, "rb") as f:
            return pickle.load(f)


class SlurmRemote(FilesystemBasedRemote):
    def __init__(self, options, my_ip):
        super(SlurmRemote, self).__init__(options, my_ip)
        self._set_world(
            int(os.environ["SLURM_PROCID"]),
            int(os.environ["SLURM_NTASKS"]),
            deployment_utils.list_as_string(os.environ["SLURM_NODELIST"]),
        )


class DALiuGERemote(FilesystemBasedRemote):
    """A remote based on DALIUGE-specific environment variables"""

    def __init__(self, options, my_ip):
        super(DALiuGERemote, self).__init__(options, my_ip)
        ips = os.environ["DALIUGE_CLUSTER_IPS"].split()
        rank = ips.index(my_ip)
        self._set_world(rank, len(ips), ips)


class DALiuGEHybridRemote(DALiuGERemote):
    """Like DALiuGERemote, but initializes MPI as well"""




#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2024
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

"""
Module that contains functionality for managing and manipulating previous sessions that have been run. 

Currently, this module only supports sessions with history stored on the _local_ filesystem. 
"""

from pathlib import Path

class PastSessionManager:
    """
    A utility class that 'manages' past sessions that have been run on the current compute
    node.

    """
    def __init__(self, work_dir: str, allow_no_graphs=False):
        self._work_dir = Path(work_dir)
        self._allow_no_graphs = allow_no_graphs

    def past_sessions(self, excluded_sessions: list = None) -> list[Path]:
        """
        Return the list of past_sessions that exists in the current working directory.

        If excluded_sessions is provided, they will be removed from the returned list.
        This is useful to remove duplications in past sessions and current session
        history when the current sessions are still tracked by the Manager.

        :params: excluded_sessions, list of sessions we want to exclude from returned
        sessions.
        """

        return [
            path for path in self._work_dir.iterdir()
            if (path.is_dir()
                and path.name not in excluded_sessions
                and self._is_session_dir(path))
        ]


    @staticmethod
    def _is_session_dir(session: Path, expected_ext=".graph") -> bool:
        """
        Inspect each file in 'session' and check if there is a file with
        'expected_suffix', which is a required file for counting that directory as a
        session directory.

        expected_ext is intended to avoid listing sessions for which there is no re-run
        or reproducibility information.
        
        :param: session, Path: a directory containing session information. 
        :param: expected_suffix: expected file extension of a file that we expect to see 
                in the session directory. If a file with this extension does not exist, 
                then this is not a 'valid' directory and we do not wish to add it to the 
                list of past sessions. 
        """

        return any(expected_ext in f.name for f in session.iterdir())





from typing import Optional
from dlg.event import EventFirer, EventHandler, Event
import pytest


class MockEventSource(EventFirer):
    def fireEvent(self, eventType, **kwargs):
        self._fireEvent(eventType, **kwargs)


class MockThrowingEventHandler(EventHandler):
    def __init__(self) -> None:
        self.wasCalled = False

    def handleEvent(self, e: Event) -> None:
        self.wasCalled = True
        raise RuntimeError("MockThrow", e)


class MockEventHandler(EventHandler):
    def __init__(self) -> None:
        self.lastEvent: Optional[Event] = None

    def handleEvent(self, e: Event) -> None:
        self.lastEvent = e


def test_listener_exception_interrupts_later_handlers():
    eventSource = MockEventSource()
    handler1 = MockEventHandler()
    handler2 = MockEventHandler()
    throwingHandler = MockThrowingEventHandler()
    eventSource.subscribe(handler1, "raise")
    eventSource.subscribe(throwingHandler, "raise")
    eventSource.subscribe(handler2, "raise")

    with pytest.raises(RuntimeError):
        eventSource.fireEvent("raise", prop="value")

    assert throwingHandler.wasCalled
    assert handler1.lastEvent is not None
    assert getattr(handler1.lastEvent, "prop") == "value"
    assert handler2.lastEvent is None



import ast
import binascii
import logging
import collections
import pickle

from dlg.ddap_protocol import DROPStates
from dlg.utils import deserialize_data
import numpy as np

from dlg.data.drops.data_base import DataDROP
import dlg.droputils as droputils
import dlg.drop_loaders as drop_loaders

from dataclasses import dataclass
from enum import Enum, IntEnum, auto
from typing import Tuple, Union

from dlg.drop import AbstractDROP

logger = logging.getLogger(f"dlg.{__name__}")

class ArgType(IntEnum):
    """
    Arguments can be positional or keyword.

    This enum is used when determining what we do with the argument, as POSITIONAL
    arguments will need to be treated differently as the order in which they appear
    is important.
    """
    POSITIONAL = auto()
    KEYWORD = auto()


class DropParser(Enum):
    RAW = "raw"
    PICKLE = "pickle"
    EVAL = "eval"
    NPY = "npy"
    DILL = "dill"
    # JSON = "json"
    PATH = "path"  # input only
    DATAURL = "dataurl"  # input only
    BINARY = "binary"
    UTF8 = "utf-8"


@dataclass()
class Argument:
    """
    Encapsulate data associated with all arguments.

    There are a few things here that will always be the default,
    but it makes more sense to have the same information always true across all instances
    of an Argument, rather than create special cases for specific dictionary elements
    """

    value: object
    encoding: DropParser = None #DropParser.DILL
    type: Union [ArgType, None] = None
    precious: bool = False
    positional: bool = False


def clean_applicationArgs(applicationArgs: dict) -> dict:
    """
    Removes arguments with None and False values, if not precious. This
    is in particular used for Bash and Docker app command lines, else
    we would have empty values for command line arguments.

    Args:
        applicationsArgs (dict): the complete set of arguments

    Returns:
        dict: a dictionary with the relevant arguments only.
    """
    cleanedArgs = {}
    if not isinstance(applicationArgs, dict):
        logger.info("applicationArgs are not passed as a dict. Ignored!")
    for name, vdict in applicationArgs.items():
        if vdict in [None, False, ""]:
            continue
        elif isinstance(vdict, bool):
            vdict = {"precious": False, "value": "", "positional": False}
        elif isinstance(vdict, dict):
            precious = vdict["precious"]
            if vdict["value"] in [None, False, ""] and not precious:
                continue
        cleanedArgs.update({name: vdict})
    logger.debug("After clean_applicationArgs: %s", cleanedArgs)
    return cleanedArgs


def serialize_applicationArgs(applicationArgs):
    """
    Unpacks the applicationArgs dictionary and returns two strings, one for
    positional arguments and one for kw arguments that can be used to construct
    the final command line.
    """
    applicationArgs = clean_applicationArgs(applicationArgs)
    pargs = []
    kwargs = {}
    for name, vdict in applicationArgs.items():
        value = vdict["value"]
        positional = vdict["positional"]
        if positional:
            pargs.append(str(value).strip())
        else:
            kwargs.update({name: value})
    logger.info("Constructed command line arguments: %s %s", pargs, kwargs)
    return pargs, kwargs

def identify_named_ports(
    port_dict: dict,
    positionalArgs: list,
    keywordArgs: dict,
    check_len: int = 0,
    mode: str = "inputs",
    parser: callable = None,
    addPositionalToKeyword: bool = False,
) -> tuple[dict, dict]:
    """
    Checks port names for matches with arguments and returns mapped ports.

    Args:
        port_dict (dict): ports {uid:name,...}
        positionalArgs (list): available positional arguments (will be modified)
        positionalPortArgs (dict): mapped arguments (will be modified)
        keywordArgs (dict): keyword arguments
        check_len (int): number of of ports to be checked
        mode (str ["inputs"]): mode, used just for logging messages
        skip_on_input (bool): skip drop if one input is skipped
        parser (function): parser function for this port
        addPositionalToKeyword (bool): Adds a positional argument to the keyword
            argument dictionary. This is useful when you have arguments
            that can be specified both positionally and as keywords. For example,
            in a Python function, you might have `func(a, b=2)`, where `a` can be
            passed positionally or as a keyword.

    Returns:
        keywordPortArgs: dict of keyword arguments from named ports

    Side effect:
        Modifies:
        - positionalArgs
        - positionalPortArgs
        - keywordArgs

    """
    # p_name = [p["name"] for p in port_dict]
    logger.debug(
        "Using named ports to remove %s from arguments port_dict: %s, check_len: %d)",
        mode,
        port_dict,
        check_len,
    )
    keywordPortArgs = {}
    positionalPortArgs = collections.OrderedDict(positionalArgs)
    positionalArgs = list(positionalArgs)
    keys = list(port_dict.keys())
    logger.debug("Checking ports: %s against %s %s", keys, positionalPortArgs, keywordArgs)
    local_parser = parser
    for i in range(check_len):
        try:
            key = port_dict[keys[i]]["name"]
            value = port_dict[keys[i]]["path"]
        except KeyError as e:
            logger.debug("portDict: %s does not have key: %s", port_dict, keys[i])
            raise KeyError from e
        logger.debug("Trying to get value for argument %s", key)
        if value is None:
            value = ""  # make sure we are passing NULL drop events
        if key in positionalArgs:
            try:
                encoding = DropParser(positionalPortArgs[key].encoding)
            except ValueError:
                logger.warning("No encoding set for %key: possible default")
                continue
            local_parser = get_port_reader_function(encoding)
            if not local_parser:
                # we prefer the port based parser if available
                local_parser = parser
            if port_dict[keys[i]]["drop"].status == DROPStates.SKIPPED:
                value = positionalPortArgs[key].value
                logger.warning("Input drop skipped! Using %s default value for parg %s", mode, key)
            elif local_parser:
                logger.debug("Reading from %s encoded port %s using %s", encoding, key, parser.__repr__())
                value = local_parser(port_dict[keys[i]]["drop"])
                try:
                    value = deserialize_data(value)
                except (TypeError, AttributeError, binascii.Error, pickle.UnpicklingError):
                    # If deserialization does not work we just
                    # stick with the value
                    pass
                positionalPortArgs[key].value = value
            logger.debug("Using %s '%s' for port %s", mode, value, key)
            positionalArgs.remove(key)
            # We have positional argument that is also a keyword
            if addPositionalToKeyword:
                keywordPortArgs.update({key: positionalPortArgs[key]})
        elif key in keywordArgs:
            try:
                encoding = DropParser(keywordArgs[key].encoding)
            except ValueError:
                logger.warning("No encoding set for %key: possible default")
                continue
            local_parser = get_port_reader_function(encoding)
            if not local_parser:
                # we prefer the port based parser if available
                local_parser = parser
            if port_dict[keys[i]]["drop"].status == DROPStates.SKIPPED:
                logger.warning("Input drop skipped! Using %s default value for parg %s", mode, key)
            if local_parser:
                logger.debug("Reading from %s encoded port using %s", encoding, parser.__repr__())
                value = local_parser(port_dict[keys[i]]["drop"])
            # if not found in appArgs we don't put them into portargs either
            # pargsDict.update({key: value})
                keywordArgs[key].value = value
            keywordPortArgs.update({key: keywordArgs[key]})
            logger.debug("Using %s of type %s for kwarg %s", mode, type(value), key)
            _ = keywordArgs.pop(key)  # remove from original arg list
        else:
            logger.debug(
                "No matching argument found for %s key %s, %s, %s",
                mode,
                key,
                keywordArgs,
                positionalArgs,
            )

    logger.debug("Returning kw mapped ports: %s", keywordPortArgs)
    return keywordPortArgs, positionalPortArgs


def check_ports_dict(ports: list) -> bool:
    """
    Checks whether all ports in ports list are of type dict. This is
    for backwards compatibility.

    Args:
        ports (list):

    Returns:
        bool: True if all ports are dict, else False
    """
    # all returns true if list is empty!
    logger.debug("Ports list is: %s", ports)
    if len(ports) > 0:
        return all(isinstance(p, dict) for p in ports)
    else:
        return False


def replace_named_ports(
    iitems: dict,
    oitems: dict,
    inport_names: dict,
    outport_names: dict,
    appArgs: dict,
    parser: callable = None,
) -> Tuple[str, str]:
    """
    Function attempts to identify CLI component arguments that match port names.

    Inputs:
        iitems: itemized input port dictionary
        oitems: itemized output port dictionary
        inport_names: dictionary of input port names (key: uid)
        outport_names: dictionary of output port names (key: uid)
        appArgs: dictionary of all arguments
        parser: reader function for ports

    This method is focused on creating two 'sets' of arguments:
    - The arguments that are passed to the application (keyword and
    positional arguments)
    - The arguments that are passed to the application that are derived from
    the ports of the drop (keywordPort and positionalPort arguments).

    Returns:
        tuple of serialized keyword arguments and positional arguments
    """
    logger.debug(
        "iitems: %s; inport_names: %s; outport_names: %s",
        iitems,
        inport_names,
        outport_names,
    )

    inputs_dict = collections.OrderedDict()
    for uid, drop in iitems:
        inputs_dict[uid] = {
            "drop": drop,
            "path": drop.path if hasattr(drop, "path") else "",
        }

    outputs_dict = collections.OrderedDict()
    for uid, drop in oitems:
        outputs_dict[uid] = {
            "drop": drop,
            "path": drop.path if hasattr(drop, "path") else "",
        }

    positionalArgs = _get_args(appArgs, positional=True)
    keywordArgs = _get_args(appArgs, positional=False)

    logger.debug(
        "posargs: %s; keyargs: %s, %s",
        positionalArgs,
        keywordArgs,
        check_ports_dict(inport_names),
    )

    # we will need an ordered dict for all positional arguments
    # thus we create it here and fill it with values
    positionalPortArgs = collections.OrderedDict(positionalArgs)
    keywordPortArgs = {}
    # This needs to be done for both the input ports and output ports on the drop.
    tmp_key, tmp_port = _process_port(
        inport_names,
        inputs_dict,
        positionalArgs,
        keywordArgs,
        iitems,
        parser,
        "inputs",
    )
    keywordPortArgs.update(tmp_key)
    positionalPortArgs.update(tmp_port)

    tmp_key, tmp_port =_process_port(
        outport_names,
        outputs_dict,
        positionalArgs,
        keywordArgs,
        oitems,
        parser,
        "outputs",
    )
    keywordPortArgs.update(tmp_key)
    positionalPortArgs.update(tmp_port)

    logger.debug("Arguments from ports: %s, %s,", keywordPortArgs, positionalPortArgs)

    # Clean arguments for Docker and Bash applications
    appArgs = clean_applicationArgs(appArgs)
    positionalArgs = _get_args(appArgs, positional=True)
    keywordArgs = _get_args(appArgs, positional=False)

    # Extract values from dictionaries - "encoding" etc. are irrelevant
    positionalArgs = {argstr: argument.value for argstr, argument in (
            positionalArgs.items())}
    keywordArgs = {argstr: argument.value for argstr, argument in keywordArgs.items()}
    keywordPortArgs = {
        argstr: argument.value for argstr, argument in keywordPortArgs.items()
    }
    positionalPortArgs = {argstr: argument.value for argstr, argument in (
        positionalPortArgs.items())}

    #  Construct the final keywordArguments and positionalPortArguments
    for k, v in keywordPortArgs.items():
        if not _is_value_empty(v):
            keywordArgs.update({k: v})
    for k, v in positionalPortArgs.items():
        if not _is_value_empty(v):
            positionalArgs.update({k: v})

    pargs = positionalArgs

    logger.debug("After port replacement: pargs: %s; keyargs: %s", pargs, keywordArgs)
    return keywordArgs, pargs

def _is_value_empty(value: object):
    """
    Check if the value is empty

    If it is a non-standard datatype, such as a numpy array, there can be truth
    ambiguities for the emptiness. In this instance, we explicitly check for the
    size of the array.

    :param value: an object that could be a primitive (int), an iterator, an object, or
    an array.
    :return: True if the value is empty.
    """
    if isinstance(value, np.ndarray):
        return True if value.size == 0 else False
    else:
        return True if value in ["", None] else False


def _process_port(
    port_names,
    ports,
    positionalArgs,
    keywordArgs,
    iitems,
    parser,
    mode,
):
    """
    For the set of port names, perform a backwards compatible update of the:

        - Keyword Arguments (application, and port name)
        - Positional Arguments (application, and port name)

    Note: This performs an IN-PLACE transformation of the dictionaries, through the
    identify_named_ports() method.
    """

    keywordPortArgs = {}
    positionalPortArgs = {}
    if check_ports_dict(port_names):
        for port in port_names:
            key = list(port.keys())[0]
            ports[key].update({"name": port[key]})
        keywordPortArgs, positionalPortArgs = identify_named_ports(
            ports,
            positionalArgs,
            keywordArgs,
            check_len=len(iitems),
            mode=mode,
            parser=parser,
        )

    else:
        for i in range(min(len(iitems), len(positionalArgs))):
            keywordPortArgs.update({list(positionalArgs)[i]: list(iitems)[i][1]})

    return keywordPortArgs, positionalPortArgs

def _get_args(appArgs, positional=False):
    """
    Separate out the arguments dependening on if we want positional or keyword style
    """
    args = {
        arg: Argument(
            value=appArgs[arg]["value"],
            encoding= appArgs[arg].get("encoding", "dill"),
            positional=positional
        )
        for arg in appArgs
        if (appArgs[arg]["positional"] == positional)
    }

    argType = "Positional" if positional else "Keyword"
    logger.debug("%s arguments: %s", argType, args)
    return args

def resolve_drop_parser(parser: Union[str, DropParser]) -> DropParser:
    """
    Resolve the drop parser to a DropParser enum value.
    """
    if isinstance(parser, str):
        try:
            parser = DropParser[parser.upper()]
        except KeyError:
            logger.critical("Invalid parser string: %s", parser)
            return DropParser.DILL
    elif not isinstance(parser, DropParser):
        logger.critical("Expected a string or DropParser enum, got %s", type(parser))
        return DropParser.DILL
    return parser

def get_port_reader_function(input_parser: DropParser):
    """
    Return the function used to read input from a named port
    """
    # Inputs are un-pickled and treated as the arguments of the function
    # Their order must be preserved, so we use an OrderedDict
    ip = None
    ip = resolve_drop_parser(input_parser)
    if ip is DropParser.PICKLE:
        # all_contents = lambda x: pickle.loads(droputils.allDropContents(x))
        reader = drop_loaders.load_pickle
    elif ip is DropParser.UTF8:
        reader = drop_loaders.load_utf8
    elif ip is DropParser.EVAL:

        def optionalEval(x):
            # Null and Empty Drops will return an empty byte string
            # which should propogate back to None
            content: str = droputils.allDropContents(x).decode("utf-8")
            logger.debug("Read %s from %s drop.", content, input_parser)
            return ast.literal_eval(content) if len(content) > 0 else None

        reader = optionalEval
    elif ip is DropParser.UTF8:

        def utf8decode(drop: "DataDROP"):
            """
            Decode utf8
            Not stored in drop_loaders to avoid cyclic imports
            """
            return droputils.allDropContents(drop).decode("utf-8")

        reader = utf8decode
    elif ip is DropParser.NPY:
        reader = drop_loaders.load_npy
    elif ip is DropParser.PATH:
        def PathFromData(x: AbstractDROP):
            # Attempt to access path from DROP "x"
            # If not Path, this could be a memory Drop with path information.
            # If so, try and decode accordingly.
            try:
                return x.path
            except AttributeError:
                return drop_loaders.load_utf8(x)
        reader = PathFromData
    elif ip is DropParser.DATAURL:
        reader = lambda x: x.dataURL
    elif ip is DropParser.DILL:
        reader = drop_loaders.load_dill
    elif ip is DropParser.BINARY:
        reader = drop_loaders.load_binary
    else:
        logger.critical("Invalid input parser specified: %s", ip.__repr__())
        return drop_loaders.load_dill
    return reader



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2019
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
from __future__ import annotations
import threading

from dlg import constants
from dlg.manager.composite_manager import DataIslandManager, MasterManager
from dlg.manager.node_manager import NodeManager
from dlg.manager.rest import NMRestServer, CompositeManagerRestServer
from dlg.utils import portIsOpen


class ManagerInfo(object):
    def __init__(self, manager, server, thread, test):
        self.manager = manager
        self.server = server
        self.thread = thread
        self.test = test

    def __enter__(self):
        pass

    def __exit__(self, *_args, **_kwargs):
        self.stop()

    def stop(self):
        self.server.stop()
        self.thread.join()
        self.manager.shutdown()
        self.test.assertFalse(self.thread.is_alive())


class ManagerStarter(object):
    def _start_manager_in_thread(
        self, port, manager_class, rest_class, *manager_args, **manager_kwargs
    ):
        manager = manager_class(*manager_args, **manager_kwargs)
        server = rest_class(manager)
        thread = threading.Thread(target=server.start, args=("localhost", port))
        thread.start()
        self.assertTrue(portIsOpen("localhost", port, 5))
        return ManagerInfo(manager, server, thread, self)

    def start_nm_in_thread(self,
                           port=constants.NODE_DEFAULT_REST_PORT,
                           events_port=constants.NODE_DEFAULT_EVENTS_PORT,
                           rpc_port=constants.NODE_DEFAULT_RPC_PORT):
        return self._start_manager_in_thread(
            port, NodeManager, NMRestServer, False, rpc_port, events_port)

            # port, NodeManager, NMRestServer, False, rpc_port, events_port)
    def start_dim_in_thread(
        self,
        nm_hosts: list[str] = None,
        port=constants.ISLAND_DEFAULT_REST_PORT,
    ):
        if not nm_hosts:
            nm_hosts = [f"localhost:{constants.NODE_DEFAULT_REST_PORT}"]
        return self._start_manager_in_thread(
            port, DataIslandManager, CompositeManagerRestServer, nm_hosts
        )

    def start_mm_in_thread(
        self,
        nm_hosts: list[str] = None,
        port=constants.MASTER_DEFAULT_REST_PORT,
    ):
        if not nm_hosts:
            nm_hosts = [f"localhost:{constants.ISLAND_DEFAULT_REST_PORT}"]
        return self._start_manager_in_thread(
            port, MasterManager, CompositeManagerRestServer, nm_hosts
        )



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2025
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

"""
Utility functions to enable submitting EAGLE graphs that are hosted on a Git repository. 
"""

import base64
import json
import requests
import sys


def gitlab_request(user, repo, branch, path):
    """
    Constructs a 'get' request using the GitLab API.

    See: https://docs.gitlab.com/api/repository_files/#get-file-from-repository

    GitLab requires URL-encoded paths, so we replace them with the %2F string for the
    user and path variables.

    :param user: The user or organisation (e.g. ICRAR/ or ska-telescope/sdp).
    :param repo: The name of the repository the file is in
    :param branch: The branch we want to get from
    :param path: The path to the file relative to the base directory of the repository
    :return:
    """
    url_enabled_user = user.replace("/", "%2F")
    url_enabled_path = path.replace("/", "%2F")
    r = requests.get(f"https://gitlab.com/api/v4/projects/{url_enabled_user}%2F"
                     f"{repo}/repository/files/{url_enabled_path}/raw?ref"
                     f"={branch}", timeout=15)
    return r.content

def github_request(user, repo, branch, path):
    """
    Constructs a 'get' request using the GitHub api

    See: https://docs.github.com/en/rest/repos/contents

    The difference here is the response content is itself a JSON dictionary, with the
    'content' element base64 encoded. We have to decode this first and then return it.

    :param user: The user or organisation (e.g. ICRAR/ or ska-telescope/sdp).
    :param repo: The name of the repository the file is in
    :param branch: The branch we want to get from
    :param path: The path to the file relative to the base directory of the repository
    :return:
    """

    request_url = f"https://api.github.com/repos/{user}/{repo}/contents/{path}?ref={branch}"

    r = requests.get(request_url,
                     timeout=15)
    content = json.loads(r.content)
    decoded_content = base64.b64decode(content['content']).decode()
    if r.status_code != 200:
        print(f"Things went wrong with {request_url}!")
        print(f"{decoded_content}")
        sys.exit(1)
    return decoded_content



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Functions and data structures to improve user-facing debugging and manage execptions
cleanly during runtime.
"""

import dill
import pickle
import logging
import dlg.exceptions as ex

from enum import Enum, auto


class ErrorCode(Enum):
    """
    Enum for Errors that occur during runtime execution on the engine.
    We use `Enum` for the auto-incremementer, and because we can use the autodoc feature
    of Sphinx to generate
    """

    DROP_ERROR = 100
    """
    An error occured during DROP initialization that was not expected. Please 
    double-check your current graph and attempt to resolve the issue based on the 
    debugging tips for DROP_ERROR codes below. 
    
    If none of the fixes address your isse, please consider opening an issue on our GitHub
    repository `here <https://github.com/ICRAR/daliuge/issues/new/choose>`_.
    """

    INCOMPLETE_DROP_SPEC = auto()
    """
    The DROP being initalized was not complete. This means that key parameters in the 
    have not been filled. If you are using a PyFuncApp, double-check you have filled in 
    both the ``func_name`` parameter if you are filling in the ``func_code`` parameter 
    too.
    """

    BAD_IMPORT = auto()
    """
    There was an issue importing the module. Consider:
    
    - Reviewing the current environment
    - Reviewing the spelling
    - Ensuring there are no missing dependencies
    """

    ENCODING_ERROR = auto()
    """
    There was an issue with reading/writing data in your drop.
    This is likely caused by using the wrong 'encoding' option when configuring
    your graph in EAGLE. Please ensure that: 
        
        - The OutputPort/InputPort on your Apps are using the same encoding
        - The Encoding matches the Type (e.g. String types are using UTF-8 encoding)
        - The DataDROP matches the right type/encoding (e.g. Using a File for its path is 
          using 'path' encoding'). 
        
    These changes will need to be added to the EAGLE graph. 
    """

    PATH_ERROR = auto()
    """
    There was an issue with the path set for this DROP, which may becaused by the 
    following: 
    
        - The file does not exist and/or you have set check_filepath_exists to be 
          True.
        - The directory does not exist at runtime
        - An environment variable that is being referred to has not been set and cannot 
          be expanded
        
    Please review the DataDROP parameter values for this DROP.    
    """

    BASH_COMMAND_FAILED = auto()
    """
    There was an issue running your BashShellApp DROP. This may be caused by: 
    
        - The application cannot find the program you are running (e.g. cannot find the
          script that is being referred to)
        - The arguments applied to the command line are incorrectly formatted; for 
          example, incorrect flags, or the wrong redirect was used. 
    """

    GRAPH_ERROR = 200
    """
    An error has occured during graph execution that was not expected. Please 
    double-check your current graph and attempt to resolve the issue based on the 
    debugging tips for SESSION_ERROR codes below. 

    If none of the fixes address your isse, please consider opening an issue on our GitHub
    repository `here <https://github.com/ICRAR/daliuge/issues/new/choose>`_.
    """

    INVALID_GRAPH_CONFIGURATION = auto()
    """
    The graph has failed to be loaded by the DALiuGE Engine. This can be caused by number 
    of issues: 

    - A python module that your graph relies on is not currently accessible in the 
      virtual environment in which DALiuGE is being run. 
    - The graph is missing key arguments - please check the JSON is complete.
    """

    SESSION_ERROR = 300
    """
    An error has occured during your session that was not expected. Please 
    double-check your current graph and attempt to resolve the issue based on the 
    debugging tips for SESSION_ERROR codes below. 
    
    If none of the fixes address your isse, please consider opening an issue on our GitHub
    repository `here <https://github.com/ICRAR/daliuge/issues/new/choose>`_ and 
    attaching the log files.
    """

    SESSION_STATE_ERROR = auto()
    """
    The session has been put in a state is was not expecting. This could be the 
    result of a synchronisation issue between the managers; consider restarting them 
    and re-running your session. 
    """

    SESSION_MISSING_ERROR = auto()
    """
    A session was not found that was requested. This typically occurs when restarting 
    the Node Manager(s) when a Data Island Manager is still running. 
    
    If a Node Manager was not restarted, please reveiw the logs and if the issue 
    persists, consider opening an issue on our GitHub repository `here 
    <https://github.com/ICRAR/daliuge/issues/new/choose>`_ and 
    attaching the log files.
    """

    @property
    def doc_url(self) -> str:
        log_message = f"Error [{self.value}] - {self.name}"
        path = (f"https://daliuge--344.org.readthedocs.build/page/debugging/errors"
                   f".html"
                f"#{__name__}.{str(self)}")
        # if os.environ.get('READTHEDOCS', None) == 'True':
        # path =  f"https://daliuge.readthedocs.org/page/errors.html#{self.name}"
        href = f"<a href={path} target='_blank' rel='noopener noreferrer'>{path}</a>'"
        return f"{log_message} occured: Please review potential issues at\n {href}"


EXCEPTION_MAP = {

    #################################################################
    # Custom DALiUGE Exceptions
    # 100 - Exceptions that inheret generic DropException
    # 200 - Exceptions that inheret GraphException
    # 300 - Exceptions that inheret SessionException
    #################################################################

    ex.InvalidDropException: ErrorCode.DROP_ERROR,
    ex.BadModuleException: ErrorCode.BAD_IMPORT,
    ex.InvalidEncodingException: ErrorCode.ENCODING_ERROR,
    ex.InvalidPathException: ErrorCode.PATH_ERROR,
    ex.IncompleteDROPSpec: ErrorCode.INCOMPLETE_DROP_SPEC,
    ex.BashAppRuntimeError: ErrorCode.BASH_COMMAND_FAILED,
    ex.InvalidGraphException: ErrorCode.GRAPH_ERROR,
    ex.IncompleteGraphError: ErrorCode.INVALID_GRAPH_CONFIGURATION,
    ex.InvalidRelationshipException: ErrorCode.INVALID_GRAPH_CONFIGURATION,
    ex.InvalidSessionException: ErrorCode.SESSION_ERROR,
    ex.InvalidSessionState: ErrorCode.SESSION_STATE_ERROR,
    ex.NoSessionException: ErrorCode.SESSION_MISSING_ERROR,

    ###################################################################
    # Standard errors that we may not catch with specific exceptions
    # based on where they might occur in the code, _but_ where we
    # are confident the cause can be resolved using our error debugging
    ###################################################################

    pickle.PickleError: ErrorCode.ENCODING_ERROR,
    dill.PickleError: ErrorCode.ENCODING_ERROR
}


def intercept_error(e: Exception):
    """
    Intercept DALiuGEExceptions during App/DataDROP runtime.

    Some exceptions we want to report and continue on running the session;
    other exceptions will leave us in an unmanageable state and thus we need
    to re-raise them to an alternative error to promote session cancellation
    prematurely.

    :param: e: The exception we have intercepted
    """

    logger = logging.getLogger(f"dlg.{__name__}")
    if type(e) != ex.ErrorManagerCaughtException:
        errorno = EXCEPTION_MAP.get(type(e), ErrorCode.DROP_ERROR)
        logger.user(errorno.doc_url)
    raise ex.ErrorManagerCaughtException from e


def manage_session_failure(func):
    """
    @decorator
    Intercept a session failure that has occured and set the session as cancelled.

    This is used to stop us in a situation where the session is in an incongruous state
    (e.g. a graph has failed to build mid-session) and we have no logic to manage this 
    in a safe way. If we detect this sort of exception has occured, we 'clean up' by 
    cancelling the session and reporting the error. 
    
    :param func: The function to which we have added this dectorator
    :return:
    """
    def manage_session(self, *args, **kwargs):
        """
        Attempt to run the decorated function 
        :param self: The session object we expect this to run within 
        :param args: 
        :param kwargs: 
        :return: If successful, the result of `func`.
        """
        try:
            return func(self, *args, **kwargs)
        except ex.InvalidSessionState:
            # sessionId = kwargs.get('sessionId')
            if args:
                session = self.sessions[next(iter(args))]
                if session:
                    session.cancel()
                return

    return manage_session
